# Copyright 2022 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import deque
from typing import Deque, Dict, Generic, Iterator, List, Optional, Set

from prezzemolo.utility import ValueType
from prezzemolo.vertex import Vertex


class Graph(Generic[ValueType]):
    def __init__(self, vertexes: Optional[List["Vertex[ValueType]"]] = None) -> None:
        self.__vertexes: Dict[Vertex[ValueType], Vertex[ValueType]] = {vertex: vertex for vertex in vertexes} if vertexes else {}
        self.__non_validated_vertexes: Set[Vertex[ValueType]] = set(vertexes) if vertexes else set()
        for vertex in self.vertexes:
            self._validate_vertex(vertex)

    def _validate_vertex(self, vertex: Vertex[ValueType]) -> bool:
        if vertex not in self.__non_validated_vertexes:
            return True
        for neighbor in vertex.neighbors:
            if neighbor not in self.__vertexes:
                return False
        self.__non_validated_vertexes.remove(vertex)
        return True

    @property
    def vertexes(self) -> Iterator[Vertex[ValueType]]:
        return iter(self.__vertexes.keys())

    def add_vertex(self, vertex: Vertex[ValueType]) -> None:
        self.__vertexes[vertex] = vertex
        self.__non_validated_vertexes.add(vertex)
        self._validate_vertex(vertex)

    def _extract_path_from_parent_dictionary(
        self, last: Vertex[ValueType], vertex_2_parent: Dict[Vertex[ValueType], Optional[Vertex[ValueType]]]
    ) -> List[Vertex[ValueType]]:
        result: List[Vertex[ValueType]] = []
        current_vertex: Optional[Vertex[ValueType]] = last
        while current_vertex:
            result.append(current_vertex)
            current_vertex = vertex_2_parent[current_vertex]
        return result

    def _breadth_first_search(
        self,
        start: Vertex[ValueType],
        end: Optional[Vertex[ValueType]],
        vertex_2_parent: Optional[Dict[Vertex[ValueType], Optional[Vertex[ValueType]]]] = None,
    ) -> bool:
        marked_set: Set[Vertex[ValueType]] = set()
        queue: Deque[Vertex[ValueType]] = deque()
        queue.append(start)
        marked_set.add(start)
        if vertex_2_parent is not None:
            vertex_2_parent[start] = None

        while len(queue) > 0:
            vertex: Vertex[ValueType] = queue.popleft()
            if vertex in self.__non_validated_vertexes and not self._validate_vertex(vertex):
                raise ValueError(f"Vertex '{vertex.name}' has a neighbor that wasn't added to the graph")
            for neighbor in vertex.neighbors:
                if not neighbor in marked_set:
                    queue.append(neighbor)
                    marked_set.add(neighbor)
                    if vertex_2_parent is not None:
                        vertex_2_parent[neighbor] = vertex
                if neighbor == end:
                    return True

        return False

    def breadth_first_search(self, start: Vertex[ValueType], end: Vertex[ValueType]) -> bool:
        # start and end are type checked inside _breadth_first_search()
        return self._breadth_first_search(start, end, vertex_2_parent=None)

    def find_shortest_path(self, start: Vertex[ValueType], end: Vertex[ValueType], reverse: bool = True) -> Optional[Iterator[Vertex[ValueType]]]:
        # start and end are type checked inside _breadth_first_search()
        vertex_2_parent: Dict[Vertex[ValueType], Optional[Vertex[ValueType]]] = {}
        if self._breadth_first_search(start, end, vertex_2_parent):
            result = self._extract_path_from_parent_dictionary(end, vertex_2_parent)
            if not reverse:
                return reversed(result)
            return iter(result)
        return None



# Copyright 2022 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys
import unittest

from prezzemolo.graph import Graph
from prezzemolo.vertex import Vertex


# pylint: disable=invalid-name
class TestGraph(unittest.TestCase):
    def setUp(self) -> None:
        self.maxDiff = None

    def test_simple_graph(self) -> None:
        v1 = Vertex[int](name="v1", data=1)
        v2 = Vertex[int](name="v2", data=2)
        v3 = Vertex[int](name="v3", data=3)
        v4 = Vertex[int](name="v4", data=4)
        v5 = Vertex[int](name="v5", data=5)
        v6 = Vertex[int](name="v6", data=6)
        v7 = Vertex[int](name="v7", data=7)

        v1.add_neighbor(v2)
        v1.add_neighbor(v3)

        v2.add_neighbor(v4)

        v3.add_neighbor(v4)

        v4.add_neighbor(v5)

        v5.add_neighbor(v2)
        v5.add_neighbor(v6)

        g1 = Graph[int]([v1, v2, v3, v4, v5, v6, v7])

        self.assertFalse(g1.breadth_first_search(v2, v7))
        self.assertFalse(g1.breadth_first_search(v2, v1))

        self.assertTrue(g1.breadth_first_search(v2, v2))
        self.assertTrue(g1.breadth_first_search(v2, v6))
        self.assertTrue(g1.breadth_first_search(v1, v5))

        path = g1.find_shortest_path(v2, v2)
        assert path
        self.assertEqual(list(path), [v2])

        path = g1.find_shortest_path(v2, v6)
        assert path
        self.assertEqual(list(path), [v6, v5, v4, v2])

        path = g1.find_shortest_path(v1, v5, reverse=False)
        assert path
        self.assertEqual(list(path), [v1, v2, v4, v5])

    def test_complex_graph(self) -> None:
        v1 = Vertex[int](name="v1", data=1)
        v2 = Vertex[int](name="v2", data=2)
        v3 = Vertex[int](name="v3", data=3)
        v4 = Vertex[int](name="v4", data=4)
        v5 = Vertex[int](name="v5", data=5)
        v6 = Vertex[int](name="v6", data=6)
        v7 = Vertex[int](name="v7", data=7)
        v8 = Vertex[int](name="v8", data=8)
        v9 = Vertex[int](name="v9", data=9)
        v10 = Vertex[int](name="v10", data=10)
        v11 = Vertex[int](name="v11", data=11)

        v1.add_neighbor(v2)
        v1.add_neighbor(v3)
        v1.add_neighbor(v4)

        v2.add_neighbor(v3)
        v2.add_neighbor(v5)

        v3.add_neighbor(v5)
        v3.add_neighbor(v6)

        v4.add_neighbor(v3)
        v4.add_neighbor(v8)

        v5.add_neighbor(v7)

        v6.add_neighbor(v5)

        v7.add_neighbor(v8)

        v8.add_neighbor(v6)
        v8.add_neighbor(v9)

        v9.add_neighbor(v11)

        v10.add_neighbor(v8)

        v11.add_neighbor(v10)

        g1 = Graph[int]([v1, v2, v3, v4, v5, v6, v7, v8, v9, v10, v11])

        self.assertFalse(g1.breadth_first_search(v11, v1))

        self.assertTrue(g1.breadth_first_search(v1, v8))

        path = g1.find_shortest_path(v1, v8, reverse=False)
        assert path
        self.assertEqual(list(path), [v1, v4, v8])

        path = g1.find_shortest_path(v4, v7)
        assert path
        self.assertEqual(list(path), [v7, v5, v3, v4])

        path = g1.find_shortest_path(v11, v1)
        self.assertIsNone(path)

    def test_graph_without_all_nodes(self) -> None:
        v1 = Vertex[int](name="v1", data=1)
        v2 = Vertex[int](name="v2", data=2)
        v3 = Vertex[int](name="v3", data=3)

        v1.add_neighbor(v2)
        v2.add_neighbor(v3)
        v3.add_neighbor(v1)

        g1 = Graph[int]([v1, v2])
        with self.assertRaisesRegex(ValueError, "Vertex .* has a neighbor that wasn't added to the graph"):
            g1.breadth_first_search(v1, v3)

        g1 = Graph[int]()
        g1.add_vertex(v1)
        g1.add_vertex(v2)
        with self.assertRaisesRegex(ValueError, "Vertex .* has a neighbor that wasn't added to the graph"):
            g1.breadth_first_search(v1, v3)

    def test_disconnected_graph(self) -> None:
        v1 = Vertex[int](name="v1", data=1)
        v2 = Vertex[int](name="v2", data=2)
        v3 = Vertex[int](name="v3", data=3)

        v1.add_neighbor(v2)

        g1 = Graph[int]([v1, v2, v3])

        self.assertFalse(g1.breadth_first_search(v1, v3))

        path = g1.find_shortest_path(v1, v3)
        self.assertIsNone(path)


if __name__ == "__main__":
    logging.basicConfig(stream=sys.stderr)
    logging.getLogger("abc").setLevel(logging.DEBUG)
    unittest.main()



# Copyright 2022 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Parametrized and extensible method to generate string representation from classes
from datetime import datetime
from decimal import Decimal
from typing import List, Optional, TypeVar

KeyType = TypeVar("KeyType", int, datetime, Decimal, float, str)  # pylint: disable=invalid-name
ValueType = TypeVar("ValueType")  # pylint: disable=invalid-name


def to_string(indent: int = 0, repr_format: bool = True, data: Optional[List[str]] = None) -> str:
    padding: str
    output: List[str] = []
    separator: str
    if not data:
        return ""

    if repr_format:
        padding = ""
        separator = ", "
        data[0] = f"{'  ' * indent}{data[0]}"
    else:
        padding = "  " * indent
        separator = "\n  "

    if data:
        for line in data:
            output.append(f"{padding}{line}")

    if repr_format:
        output[-1] += ")"

    # Joining by separator adds one level of indentation to internal fields (like id) in str mode, which is correct.
    return separator.join(output)



# Copyright 2022 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Callable, Dict, Generic, Iterator, List, Optional
from prezzemolo.utility import ValueType

from prezzemolo.utility import to_string


class Vertex(Generic[ValueType]):
    def __init__(self, name: str, data: Optional[ValueType] = None) -> None:
        self.__name: str = name
        self.__data: Optional[ValueType] = data
        self.__neighbors: Dict[Vertex[ValueType], Vertex[ValueType]] = {}
        self.__edge_weights: Dict[Vertex[ValueType], float] = {}

    def to_string(self, indent: int = 0, repr_format: bool = True, extra_data: Optional[List[str]] = None) -> str:
        class_specific_data: List[str] = []
        stringify: Callable[[object], str] = repr if repr_format else str  # type: ignore[assignment]
        if repr_format:
            class_specific_data.append(f"{type(self).__name__}(name={stringify(self.name)}")
        else:
            class_specific_data.append(f"{type(self).__name__}:")
            class_specific_data.append(f"name={stringify(self.name)}")
        class_specific_data.append(f"data={stringify(self.data)}")
        class_specific_data.append(f"neighbors=[{stringify(', '.join([neighbor.name for neighbor in self.neighbors]))}]")
        class_specific_data.append(f"weights={stringify({neighbor.name:weight for neighbor, weight in self.__edge_weights.items()})}")

        if extra_data:
            class_specific_data.extend(extra_data)

        return to_string(indent=indent, repr_format=repr_format, data=class_specific_data)

    def __str__(self) -> str:
        return self.to_string(indent=0, repr_format=False)

    def __repr__(self) -> str:
        return self.to_string(indent=0, repr_format=True)

    def __eq__(self, other: object) -> bool:
        if not other:
            return False
        if not isinstance(other, Vertex):
            raise TypeError(f"Operand has non-Vertex value: {repr(other)}")
        return self.name == other.name

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash(self.name)

    @property
    def name(self) -> str:
        return self.__name

    @property
    def data(self) -> Optional[ValueType]:
        return self.__data

    @property
    def neighbors(self) -> Iterator["Vertex[ValueType]"]:
        return iter(self.__neighbors.keys())

    def add_neighbor(self, vertex: "Vertex[ValueType]", weight: float = 0.0) -> None:
        if vertex in self.__neighbors:
            raise ValueError(f"Vertex '{vertex.name}' has already been added to vertex '{self.name}'")
        self.__neighbors[vertex] = vertex
        if weight != 0.0:
            self.__edge_weights[vertex] = weight

    def add_neighbor_bidirectional(self, vertex: "Vertex[ValueType]", weight: float = 0.0) -> None:
        self.add_neighbor(vertex, weight)
        vertex.add_neighbor(self, weight)

    def get_weight(self, neighbor: "Vertex[ValueType]") -> float:
        return self.__edge_weights[neighbor] if neighbor in self.__edge_weights else 0.0
